use clichess;
use clichess::Player;
use clichess::RecvPositionError;
use clichess::UserRole;
use serde_json::Value;
use shakmaty::{fen, Chess, Color, Setup};
use std::collections::HashMap;
use std::fs;
use std::os::unix::net::{UnixListener, UnixStream};
use std::sync::{Arc, Condvar, Mutex};

use std::thread;

#[derive(Clone)]
struct Server {
    id: usize,
    chess_position: Arc<Mutex<Chess>>,
    players: Arc<Mutex<HashMap<usize, (Player, Arc<(Mutex<String>, Condvar)>)>>>,
    receiving_buffer: Arc<(Mutex<String>, Condvar)>,
}

fn main() {
    let chess = Arc::new(Mutex::new(Chess::default()));
    let players = Arc::new(Mutex::new(HashMap::new()));
    let mut counter = 0;
    fs::remove_file("/tmp/clichess.socket");

    let listener = UnixListener::bind("/tmp/clichess.socket").unwrap();

    // accept connections and process them, spawning a new thread for each one
    for stream in listener.incoming() {
        match stream {
            Ok(stream) => {
                let server = Server {
                    id: counter,
                    chess_position: chess.clone(),
                    players: players.clone(),
                    receiving_buffer: Arc::new((Mutex::new(String::new()), Condvar::new())),
                };
                /* connection succeeded */
                thread::spawn(move || handle_player(stream, server));
                counter += 1;
            }
            Err(err) => {
                println!("Could not create a connection: {}", err);
                break;
            }
        }
    }
}

fn handle_player(mut stream: UnixStream, server: Server) {
    match initialize_client(&mut stream, &server) {
        Ok((player, player_turn)) => main_loop(&mut stream, &server, player, player_turn),
        Err(e) => println!("User id {} could not be initialized: {}", server.id, e),
    };
    player_disconnected(&server);
}

fn initialize_client(
    stream: &mut UnixStream,
    server: &Server,
) -> Result<(Player, Color), RecvPositionError> {
    //create player
    let player = create_player(server, stream)?;
    let player_turn: Color;
    //send current position to the player
    println!(
        "server {}, send current position to the player...",
        server.id
    );
    {
        let chess = server.chess_position.lock().unwrap();
        player_turn = chess.turn();
        clichess::write_to_stream(stream, fen::fen(&*chess)).unwrap();
    }
    println!("server {}, current position to the player sent", server.id);
    Ok((player, player_turn))
}

fn main_loop(stream: &mut UnixStream, server: &Server, player: Player, mut player_turn: Color) {
    loop {
        if clichess::is_player_turn(&player, player_turn) {
            //let go of the lock while waiting for user input.
            println!("server {}, waiting for player move..", server.id);
            let input;
            match clichess::read_line_from_stream(stream) {
                Ok(i) => input = i,
                Err(e) => {
                    println!("Error while getting user input: {}", e);
                    break;
                }
            };
            if input == "exit" {
                break;
            }
            {
                let mut chess = server.chess_position.lock().unwrap();
                let players = server.players.lock().unwrap();
                println!("User tried to play {}", input);
                match clichess::try_to_play_move(&chess, input) {
                    Ok(played_chess) => *chess = played_chess,
                    Err(e) => println!("Error: {}", e),
                };
                let chessfen = fen::fen(&*chess);
                for (id, (_, cvar_buffer)) in players.iter() {
                    if server.id != *id {
                        cvar_buffer.0.lock().unwrap().push_str(&chessfen);
                        cvar_buffer.1.notify_one();
                    }
                }
                clichess::write_to_stream(stream, chessfen).unwrap();
                player_turn = chess.turn();
            }
        } else {
            {
                let (buffer_mutex, cvar) = &*server.receiving_buffer;
                let mut buffer = buffer_mutex.lock().unwrap();
                println!("server id: {}, waiting for move to send...", server.id);
                while buffer.is_empty() {
                    buffer = cvar.wait(buffer).unwrap();
                }
                println!("server id: {}, sending {}", server.id, buffer);
                match clichess::write_to_stream(stream, buffer.clone()) {
                    Ok(()) => buffer.clear(),
                    Err(e) => {
                        println!("{}", e);
                        break;
                    }
                }
                let chess = server.chess_position.lock().unwrap();
                player_turn = chess.turn();
            }
        }
    }
}

fn create_player(server: &Server, stream: &mut UnixStream) -> Result<Player, RecvPositionError> {
    println!("Creating player {}...", server.id);
    //get player name and pubkey
    let username_pubkey_json =
        clichess::read_line_from_stream(stream).expect("Player closed connection.");
    let username_pubkey_value: Value = serde_json::from_str(&username_pubkey_json).unwrap();
    let username = username_pubkey_value["username"].to_string();
    let public_key = username_pubkey_value["pubkey"].to_string();
    println!("got username: {}", username);
    println!("got pubkey: {}", public_key);
    let role = receive_user_role(server, stream)?;
    let mut players = server.players.lock().unwrap();
    let player = Player {
        role,
        username,
        public_key,
    };
    players.insert(server.id, (player.clone(), server.receiving_buffer.clone()));
    println!("Created player {}", server.id);
    Ok(player)
}

fn receive_user_role(
    server: &Server,
    stream: &mut UnixStream,
) -> Result<UserRole, RecvPositionError> {
    let mut chosen_role: UserRole;
    loop {
        //send available roles
        let available_roles = compute_available_roles_to_str(server);
        println!("Computed available_roles as str: {}", available_roles);
        clichess::write_to_stream(stream, available_roles.clone()).unwrap();
        //receive chosen role
        let chosen_role_str =
            clichess::read_line_from_stream(stream).expect("Player closed connection.");
        if chosen_role_str == "exit" {
            return Err(RecvPositionError::UserCanceledError);
        }
        println!("Client id {} has chosen {}", server.id, chosen_role_str);
        let chosen_role_parse = clichess::parse_to_role(&chosen_role_str);

        if chosen_role_parse.is_err() {
            return Err(RecvPositionError::ParsePositionError);
        }
        chosen_role = chosen_role_parse.unwrap();
        //check if role is still available after the choice
        {
            let mut available_roles_after_choice =
                vec![UserRole::White, UserRole::Black, UserRole::Spectator];
            let players = server.players.lock().unwrap();
            for (_, (player, _)) in players.iter() {
                match available_roles_after_choice
                    .iter()
                    .position(|r| *r == player.role)
                {
                    Some(index) => available_roles_after_choice.remove(index),
                    None => continue,
                };
            }
            if available_roles_after_choice.contains(&chosen_role) {
                println!("OK");
                clichess::write_to_stream(stream, String::from("OK")).unwrap();
                clichess::read_line_from_stream(stream).expect("Player closed connection.");
                break;
            } else {
                println!("KO");
                clichess::write_to_stream(stream, String::from("KO")).unwrap();
                continue;
            }
        }
    }
    Ok(chosen_role)
}

fn compute_available_roles_to_str(server: &Server) -> String {
    let mut available_roles = vec![UserRole::White, UserRole::Black, UserRole::Spectator];
    let players = server.players.lock().unwrap();
    for (_, (player, _)) in players.iter() {
        match available_roles.iter().position(|r| *r == player.role) {
            Some(index) => available_roles.remove(index),
            None => continue,
        };
    }
    let available_roles_str: Vec<String> = available_roles
        .iter()
        .map(|r| clichess::role_to_str(r))
        .collect();
    available_roles_str.join(",")
}

fn player_disconnected(server: &Server) {
    let mut players = server.players.lock().unwrap();
    players.remove(&server.id);
    println!("Player disconnected.");
}
